#!/usr/bin/env sage
#
# Usage: sage ./maximal_etf_certificate.sage 7

import collections
import itertools
import os
import sys

from sage.rings.integer import Integer
from sage.rings.rational_field import QQ
from sage.graphs.strongly_regular_db import strongly_regular_graph
from sage.matrix.special import identity_matrix, ones_matrix, zero_matrix
from sage.matrix.constructor import matrix
from sage.functions.other import sqrt
from sage.functions.generalized import sign
from sage.misc.sage_eval import sage_eval


_VALID_DIMENSIONS = (2, 3, 7,)
_CENTERED_REPEATED_VALUE_FACTORS = {2: 3, 3: 3, 7: 7 / 2}

r = Integer(sage_eval(sys.argv[1]))
assert r in _VALID_DIMENSIONS

N = Integer(r * (r + 1) / 2)

if r == 3:
    from sage.rings.number_field.number_field import QuadraticField
    R.<sqrt5> = QuadraticField(5)
    mu = 1 / sqrt5
else:
    R = QQ
    mu = sqrt((N - r) / (r * (N - 1)))


#
# Build ETF Gram matrix
#

if r == 2:
    Z = matrix(R, [
        [1, -1/2, -1/2],
        [-1/2, 1, -1/2],
        [-1/2, -1/2, 1]])
else:
    A = strongly_regular_graph(
        N - 1,
        N / 2 - 1 + (N / (2 * r) - 1) / mu,
        (3 * (N / 2 - 1 + (N / (2 * r) - 1) / mu) - (N - 1) - 1) / 2,
        (N / 2 - 1 + (N / (2 * r) - 1) / mu) / 2
    ).adjacency_matrix()

    Z = identity_matrix(R, N)
    Z[1:, 0] = mu
    Z[0, 1:] = mu
    Z[1:, 1:] = (
        2 * mu * A +
        (mu + 1) * identity_matrix(R, N - 1) -
        mu * ones_matrix(R, N - 1, N - 1))


#
# Build polynomial matrix
#

# Value for (ii) (jk)
skewed_repeated_value = 1 / (N * N * (N - 1)) * 3 / 2
# Value for (ij) (ik)
centered_repeated_value = (
    skewed_repeated_value * _CENTERED_REPEATED_VALUE_FACTORS[r])
# Value for (ij) (ij)
main_diag_value = skewed_repeated_value * N
# Value for (ii) (jj)
diag_minor_value = (1 - N * (N - 1) / 2 * main_diag_value) / N / N

subsets = [list(t) for t in itertools.combinations_with_replacement(range(N), 2)]
X = zero_matrix(R, len(subsets), len(subsets), sparse=True)

print 'Total subsets:', len(subsets)

deg_2_terms_dict = collections.defaultdict(itertools.repeat(R(0)).next)
constant_term = R(0)

for i, s1 in enumerate(subsets):
    for j, s2 in enumerate(subsets):
        ix_counts = collections.defaultdict(int)
        ix_counts[s1[0]] += 1
        ix_counts[s1[1]] += 1
        ix_counts[s2[0]] += 1
        ix_counts[s2[1]] += 1

        num_unique = len(ix_counts)

        if num_unique == 4:
            X[i, j] = R(0)
        elif num_unique == 3:
            repeated = [k for k, v in ix_counts.iteritems() if v == 2][0]
            non_repeateds = [k for k, v in ix_counts.iteritems() if v == 1]

            # x_i^2 * x_jx_k -> 2 * 2 = 4 spots in non-collapsed matrix
            if s1[0] == s1[1] or s2[0] == s2[1]:
                if Z[non_repeateds[0], non_repeateds[1]] > 0:
                    X[i, j] = -skewed_repeated_value
                else:
                    X[i, j] = skewed_repeated_value
            # x_ix_j * x_ix_k -> (4 2) * 2 = 12 spots in non-collapsed matrix
            else:
                ixs = ix_counts.keys()
                num_edges = 0
                if Z[ixs[0], ixs[1]] > 0:
                    num_edges += 1
                if Z[ixs[0], ixs[2]] > 0:
                    num_edges += 1
                if Z[ixs[1], ixs[2]] > 0:
                    num_edges += 1

                if num_edges == 0:  # No edges
                    X[i, j] = centered_repeated_value
                elif num_edges % 2 == 1:
                    X[i, j] = R(0)
                elif (Z[repeated, non_repeateds[0]] > 0 and
                      Z[repeated, non_repeateds[1]] > 0):
                    X[i, j] = centered_repeated_value
                else:
                    X[i, j] = -centered_repeated_value

            deg_2_terms_dict[frozenset(non_repeateds)] += X[i, j]
        elif num_unique == 2:
            # x_i^2 * x_j^2
            if s1[0] == s1[1] and s2[0] == s2[1]:
                X[i, j] = diag_minor_value
                constant_term += X[i, j]
            # x_ix_j * x_ix_j
            elif s1[0] == s2[0] and s1[1] == s2[1]:
                X[i, j] = main_diag_value
                constant_term += X[i, j]
            # x_i^2 * x_ix_j
            else:
                repeated = [k for k, v in ix_counts.iteritems() if v == 3][0]
                non_repeated = [k for k, v in ix_counts.iteritems() if v == 1][0]

                if Z[repeated, non_repeated] > 0:
                    X[i, j] = -skewed_repeated_value
                else:
                    X[i, j] = skewed_repeated_value

                term_key = frozenset([repeated, non_repeated])
                deg_2_terms_dict[term_key] += X[i, j]
        elif num_unique == 1:
            X[i, j] = diag_minor_value
            constant_term += X[i, j]

print ''
print 'Built X'
print ''


#
# Verify positive semidefinitess
#

column_space = X.column_space()
column_space_basis = column_space.basis_matrix()
X_trunc = column_space_basis * X * column_space_basis.T

print ''
try:
    X_trunc.cholesky()
    print 'Verified X >> 0'
except e:
    print 'Cholesky decomposition failed'
    exit
print ''


#
# Check polynomial structure
#

deg_2_coeffs = deg_2_terms_dict.values()
for i in range(len(deg_2_coeffs) - 1):
    assert abs(deg_2_coeffs[i]) == abs(deg_2_coeffs[i + 1])

for term_key, coeff in deg_2_terms_dict.iteritems():
    ix_1, ix_2 = list(term_key)
    assert sign(coeff) == -sign(Z[ix_1, ix_2])

absolute_coeff = abs(deg_2_coeffs[0])

rhs_value = constant_term / absolute_coeff

print skewed_repeated_value / absolute_coeff
print centered_repeated_value / absolute_coeff
print main_diag_value / absolute_coeff
print diag_minor_value / absolute_coeff
print '--'
print rhs_value

print 'Verified inequality:'
print '  \sum_{1 \leq i < j \leq N} \sgn(Z_{ij})x_ix_j \leq', rhs_value
